package io.provis.assembly;

import io.provis.assembly.action.artifact.ArtifactMetadataGleaner;
import io.provis.assembly.action.artifact.WriteToDiskAction;
import io.provis.assembly.model.ProvisoArtifactMetadata;
import io.provis.model.Action;
import io.provis.model.ArtifactSet;
import io.provis.model.ProvisoArtifact;
import io.provis.provision.Provisioner;
import io.provis.provision.ProvisioningRequest;
import io.provis.provision.ProvisioningResult;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;

import com.google.common.collect.Maps;

@Named
@Singleton
public class DefaultRuntimeAssembler implements RuntimeAssembler {

  private Provisioner provisioner;

  private ArtifactMetadataGleaner artifactMetadataGleaner;

  @Inject
  public DefaultRuntimeAssembler(Provisioner provisioner, ArtifactMetadataGleaner artifactMetadataGleaner) {
    this.provisioner = provisioner;
    this.artifactMetadataGleaner = artifactMetadataGleaner;
  }

  public ProvisioningResult assemble(ProvisioningRequest request) {

    //
    // We probably want to make sure all the operations can be done first
    //
    
    for (ArtifactSet fileSet : request.getRuntimeAssembly().getFileSets()) {
      try {
        processFileSet(request, fileSet);
      } catch (Exception e) {
        //
        // Handle
        //
      }
    }

    try {
      processRuntimeActions(request);
    } catch (Exception e) {
      //
      // Handle
      //
    }

    return new ProvisioningResult(request.getRuntimeAssembly());
  }

  private void processFileSet(ProvisioningRequest request, ArtifactSet fileSet) throws Exception {

    resolveFileSetOutputDirectory(request, fileSet);
    resolveFileSetArtifacts(request, fileSet);
    processArtifactsWithActions(fileSet.getResolvedArtifacts().values(), fileSet.getActualOutputDirectory());
    processFileSetActions(fileSet.getActualOutputDirectory(), fileSet);

    if (fileSet.getFileSets() != null) {
      for (ArtifactSet childFileSet : fileSet.getFileSets()) {
        processFileSet(request, childFileSet);
      }
    }
  }

  private void resolveFileSetOutputDirectory(ProvisioningRequest request, ArtifactSet fileSet) {
    ArtifactSet parent = fileSet.getParent();
    if (parent != null) {
      fileSet.setActualOutputDirectory(new File(parent.getActualOutputDirectory(), fileSet.getDirectory()));
    } else {
      fileSet.setActualOutputDirectory(new File(request.getOutputDirectory(), fileSet.getDirectory()));
    }
  }

  private void resolveFileSetArtifacts(ProvisioningRequest request, ArtifactSet fileSet) {
    //
    // Set Parent = [a, b, c]
    // Set Child = [a, b, c, d, e, f]
    //
    // First we want to collect all the dependencies that a FileSet may yield.
    //
    // We want to use this first in a calculation of overlapping dependencies between FileSets that have a parent-->child relationship. We are making the assumption that a
    // classloader relationship will be setup along the lines of the parent-->child relationship. So we only want to place in the child's directory the artifacts that
    // are not present in the parent.
    //
    Map<String, ProvisoArtifact> artifacts = provisioner.resolveFileSet(request, fileSet);
    ArtifactSet parent = fileSet.getParent();
    if (parent != null) {
      Map<String, ProvisoArtifact> parentArtifacts = fileSet.getParent().getResolvedArtifacts();
      Map<String, ProvisoArtifact> resolved = Maps.difference(artifacts, parentArtifacts).entriesOnlyOnLeft();
      fileSet.setResolvedArtifacts(resolved);
    } else {
      fileSet.setResolvedArtifacts(artifacts);
    }
  }

  //
  // Process actions that apply across the entire runtime installation
  //
  private void processRuntimeActions(ProvisioningRequest request) throws Exception {
    for (Action action : request.getRuntimeAssembly().getActions()) {
      action.execute();
    }
  }

  //
  // Process actions that apply across filesets
  //
  private void processFileSetActions(File outputDirectory, ArtifactSet fileSet) throws Exception {

    for (Action action : fileSet.getActions()) {
      action.execute();
    }
  }

  //
  // Process actions that apply to artifacts
  //
  private void processArtifactsWithActions(Collection<ProvisoArtifact> artifacts, File outputDirectory) throws Exception {

    for (ProvisoArtifact artifact : artifacts) {
      //
      // Find any actions specified in the metadata of the artifact
      //
      ProvisoArtifactMetadata artifactMetadata = artifactMetadataGleaner.gleanMetadata(artifact);
      List<Action> mergedArtifactActions = mergeActionsAndSetDefaults(artifact, artifactMetadata, outputDirectory);

      for (Action action : mergedArtifactActions) {
        action.execute();
      }

      if (artifactMetadata != null) {
        if (artifactMetadata.getDirectoryActions() != null) {
          for (Action action : artifactMetadata.getArtifactActions()) {
            action.execute();
          }
        }
      }
    }
  }

  private List<Action> mergeActionsAndSetDefaults(ProvisoArtifact artifact, ProvisoArtifactMetadata artifactMetadata, File outputDirectory) {

    List<Action> actions = new ArrayList<Action>();

    for (Action action : artifact.getActions()) {
      actions.add(action);
    }

    if (artifactMetadata != null && artifactMetadata.getArtifactActions() != null) {
      for (Action action : artifactMetadata.getArtifactActions()) {
        actions.add(action);
      }
    }

    //
    // If no other actions have been set by default we want to write the artifact to disk.
    //
    if (actions.isEmpty()) {
      actions.add(new WriteToDiskAction(artifact, outputDirectory));
    }    

    return actions;
  }
}
