package io.provis.provision;

import io.provis.model.ArtifactSet;
import io.provis.model.ProvisoArtifact;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;

import org.apache.maven.repository.internal.DefaultServiceLocator;
import org.apache.maven.repository.internal.MavenRepositorySystemSession;
import org.apache.maven.settings.Mirror;
import org.apache.maven.settings.Settings;
import org.apache.maven.settings.building.DefaultSettingsBuilderFactory;
import org.apache.maven.settings.building.DefaultSettingsBuildingRequest;
import org.apache.maven.settings.building.SettingsBuilder;
import org.apache.maven.settings.building.SettingsBuildingException;
import org.apache.maven.settings.building.SettingsBuildingRequest;
import org.slf4j.Logger;
import org.sonatype.aether.RepositorySystem;
import org.sonatype.aether.RepositorySystemSession;
import org.sonatype.aether.artifact.Artifact;
import org.sonatype.aether.artifact.ArtifactType;
import org.sonatype.aether.collection.CollectRequest;
import org.sonatype.aether.connector.async.AsyncRepositoryConnectorFactory;
import org.sonatype.aether.graph.Dependency;
import org.sonatype.aether.repository.Authentication;
import org.sonatype.aether.repository.LocalRepository;
import org.sonatype.aether.repository.Proxy;
import org.sonatype.aether.repository.RemoteRepository;
import org.sonatype.aether.resolution.ArtifactResult;
import org.sonatype.aether.resolution.DependencyRequest;
import org.sonatype.aether.resolution.DependencyResolutionException;
import org.sonatype.aether.resolution.DependencyResult;
import org.sonatype.aether.spi.connector.RepositoryConnectorFactory;
import org.sonatype.aether.util.artifact.DefaultArtifactType;
import org.sonatype.aether.util.filter.ExclusionsDependencyFilter;

@Named
@Singleton
public class DefaultProvisioner implements Provisioner {

  @Inject
  private Logger logger;

  public Map<String, ProvisoArtifact> resolveFileSet(ProvisioningRequest provisioningRequest, ArtifactSet fileSet) {

    provisioningRequest = validateReasonableDefaultsForProvisioningRequest(provisioningRequest);
    
    CollectRequest request = new CollectRequest();
    request.setRepositories(provisioningRequest.getRemoteRepositories());
    //
    // Resolve versions
    //
    Map<String, ProvisoArtifact> artifacts = resolveVersions(fileSet.getArtifactMapKeyedByGA(), provisioningRequest.getVersionMap());
    for (ProvisoArtifact artifact : artifacts.values()) {
      //
      // <groupId>:<artifactId>[:<extension>[:<classifier>]]:<version>
      //
      ArtifactType type;
      if (artifact.getExtension().equals("tar.gz")) {
        type = new DefaultArtifactType("tar.gz", "tar.gz", "", "java", false, true);
      } else if (artifact.getExtension().equals("war")) {
        type = new DefaultArtifactType("war", "war", "", "java", false, true);
      } else {
        type = new DefaultArtifactType("jar", "jar", "", "java");
      }

      artifact = (ProvisoArtifact) artifact.setProperties(type.getProperties());
      Dependency dependency = new Dependency(artifact, "runtime");
      request.addDependency(dependency);
    }

    //
    // Add an exclude filter if necessary
    //
    DependencyRequest dependencyRequest = new DependencyRequest(request, null);
    if (fileSet.getExcludes() != null) {
      dependencyRequest.setFilter(new ExclusionsDependencyFilter(fileSet.getExcludes()));
    }

    //
    // Treat the parent's resolved artifacts as set of managed dependencies for the child
    //
    if (fileSet.getParent() != null) {
      for (Artifact artifact : fileSet.getParent().getResolvedArtifacts().values()) {
        request.addManagedDependency(new Dependency(artifact, "runtime"));
      }
    }

    DependencyResult dependencyResult;
    try {
      dependencyResult = provisioningRequest.getRepositorySystem().resolveDependencies(provisioningRequest.getResolutionSession(), dependencyRequest);
    } catch (DependencyResolutionException e) {
      throw new ProvisioningException(e.getMessage(), e);
    }

    Map<String, ProvisoArtifact> artifactMapKeyedByGa = new HashMap<String, ProvisoArtifact>();
    for (ArtifactResult artifactResult : dependencyResult.getArtifactResults()) {
      Artifact a = artifactResult.getArtifact();
      String ga = a.getGroupId() + ":" + a.getArtifactId();
      if (a instanceof ProvisoArtifact) {
        artifactMapKeyedByGa.put(ga, (ProvisoArtifact) a);
      } else {
        artifactMapKeyedByGa.put(ga, new ProvisoArtifact(a));
      }
    }

    fileSet.setResolvedArtifacts(artifactMapKeyedByGa);

    return artifactMapKeyedByGa;
  }

  private Map<String, ProvisoArtifact> resolveVersions(Map<String, ProvisoArtifact> artifacts, Map<String, String> versionMap) {
    Map<String, ProvisoArtifact> list = new LinkedHashMap<String, ProvisoArtifact>();
    for (ProvisoArtifact artifact : artifacts.values()) {
      if (artifact.getVersion().equals("NOTSET")) {
        String version = versionMap.get(artifact.getGroupId() + ":" + artifact.getArtifactId());
        artifact = (ProvisoArtifact) artifact.setVersion(version);
      }
      list.put(artifact.getGA(), artifact);
    }
    return list;
  }  
  
  //
  // Population of the ProvisioningRequest with default values if they are not set. We default to Maven default.
  //  
  
  private ProvisioningRequest validateReasonableDefaultsForProvisioningRequest(ProvisioningRequest request) {

    if (request.getRemoteRepositories() == null) {
      List<RemoteRepository> remoteRepositories = new ArrayList<RemoteRepository>();
      remoteRepositories.add(getRemoteRepository(new Properties()));
      request.setRemoteRepositories(remoteRepositories);
    }

    if (request.getRepositorySystem() == null) {
      request.setRepositorySystem(getRepositorySystem());
    }

    if (request.getResolutionSession() == null) {
      request.setLocalRepository(getLocalRepository(new Properties()));
      request.setResolutionSession(getRepositorySystemSession(request));
    }
    
    return request;
  }  
  
  public RepositorySystem getRepositorySystem() {
    DefaultServiceLocator serviceLocator = new DefaultServiceLocator();
    serviceLocator.addService(RepositoryConnectorFactory.class, AsyncRepositoryConnectorFactory.class);
    return serviceLocator.getService(RepositorySystem.class);
  }

  private RepositorySystemSession getRepositorySystemSession(ProvisioningRequest request) {
    final MavenRepositorySystemSession repoSession = new MavenRepositorySystemSession();
    repoSession.setLocalRepositoryManager(request.getRepositorySystem().newLocalRepositoryManager(new LocalRepository(request.getLocalRepository())));
    repoSession.setTransferListener(new AetherTransferListener());
    return repoSession;
  }

  public String getLocalRepository(Properties properties) {

    String localRepository = properties.getProperty("repo.local");
    if (localRepository == null || localRepository.length() <= 0) {
      Settings settings = readMavenSettings();
      localRepository = settings.getLocalRepository();
      if (localRepository == null || localRepository.length() <= 0) {
        localRepository = new File(System.getProperty("user.home"), ".m2/repository").getAbsolutePath();
      }
    }

    return localRepository;
  }

  public RemoteRepository getRemoteRepository(Properties properties) {
    String user = System.getProperty("user.name");
    String repoUrl = properties.getProperty("repo.url", "http://repo1.maven.org/maven2/");
    String repoUser = properties.getProperty("repo.username", user);
    String repoPass = properties.getProperty("repo.password");
    //
    // Proxy
    //
    String proxyHost = properties.getProperty("proxy.host", "");
    String proxyPort = properties.getProperty("proxy.port", "80");
    String proxyUser = properties.getProperty("proxy.username", user);
    String proxyPass = properties.getProperty("proxy.password");

    String repositoryId = "proviso";

    Settings settings = readMavenSettings();
    //
    // If there is a mirrorOf in the settings settings
    if (settings.getMirrors() != null && settings.getMirrors().size() == 1) {
      Mirror mirror = settings.getMirrors().get(0);
      repoUrl = mirror.getUrl();
      repositoryId = mirror.getId();
      //
      // account for having no username or password
      //
      if (settings.getServer(repositoryId) != null) {
        repoUser = settings.getServer(repositoryId).getUsername();
        repoPass = settings.getServer(repositoryId).getPassword();
      }
    }

    RemoteRepository remoteRepository = new RemoteRepository(repositoryId, "default", repoUrl);
    remoteRepository.setAuthentication(new Authentication(repoUser, repoPass));
    if (proxyHost.length() > 0) {
      int port;
      try {
        port = Integer.parseInt(proxyPort);
      } catch (NumberFormatException e) {
        throw new ProvisioningException("Invalid proxy port " + proxyPort + ": " + e.getMessage(), e);
      }
      remoteRepository.setProxy(new Proxy(Proxy.TYPE_HTTP, proxyHost, port, new Authentication(proxyUser, proxyPass)));
    }
    return remoteRepository;
  }

  public Settings readMavenSettings() {
    SettingsBuilder settingsBuilder = new DefaultSettingsBuilderFactory().newInstance();
    SettingsBuildingRequest request = new DefaultSettingsBuildingRequest();
    request.setSystemProperties(System.getProperties());
    request.setUserSettingsFile(new File(System.getProperty("user.home"), ".m2/settings.xml"));

    Settings settings;
    try {
      settings = settingsBuilder.build(request).getEffectiveSettings();
    } catch (SettingsBuildingException e) {
      settings = new Settings();
    }

    return settings;
  }  
}
